import pandas as pd
import tfgen.utils as utils


def get_observable_ec(attributes):
    """
    This function finds all observable event classes from an offline dataset.
    :param attributes:
    :return:
    """
    # check if attributes is a dataframe
    if isinstance(attributes, pd.DataFrame):
        attributes = list(attributes.astype(str).values)
    else:
        attributes = list(attributes)

    event_classes = set([utils.convert_attr_to_ec(attribute) for attribute in attributes])
    return sorted(list(event_classes))



from abc import ABC

from tfgen.methods.base_method import BaseMethod
import tfgen.utils as utils
from time import time
from tfgen import const
import numpy as np


class Classic(BaseMethod, ABC):
    def __init__(self, ec_lookup_table, window_size, input_stream, output_stream):
        super().__init__(ec_lookup_table, window_size, input_stream, output_stream)

        # 10% performance increase by increase/decrease the transition_matrix by a small step (1/window size)
        # instead of normalise like transition_count/window size. Saved a division operation.
        self.step = 1 / window_size

    def init_window(self):
        for _ in range(self.window_size):
            data = self.get_next_data()

            if self.finished:
                return

            self.process_a_single_event(data)

        self._output_stream.put(
            (self.current_case, np.array(self.transition_counts))
        )

    def process_a_single_event(self, data_tuple):
        case_id, event_attr = data_tuple

        if all(map(lambda x: x == const.TOKEN_END_OF_TRACE, event_attr)):
            event_class = const.TOKEN_END_OF_TRACE
        else:
            event_class = utils.convert_attr_to_ec(event_attr)

        prev = self.tet.update_event(case_id, event_class)

        # reduce the count for transition that went outside the window.
        if len(self.sliding_window_buffer) >= self.window_size:
            transition_out_window = self.sliding_window_buffer[0]
            self.transition_counts[transition_out_window] -= self.step

        # add the count for the transition that come in.
        transition_new = (self.index_lookup(prev), self.index_lookup(event_class))
        self.sliding_window_buffer.append(transition_new)
        self.transition_counts[transition_new] += self.step

        self.current_case = case_id

    def start_processing(self):
        self.init_window()
        if self.finished:
            return

        count = 0

        last_benchmark = time()

        while True:
            data = self.get_next_data()

            if self.finished:
                return

            self.process_a_single_event(data)
            count += 1

            if not count % 100000:
                now = time()
                print('Processing speed: {} events/s. TET size: {}'.format(
                    100000 / (now - last_benchmark),
                    len(self.tet.table)
                ))
                last_benchmark = now
                count = 0

            self._output_stream.put(
                (self.current_case, np.array(self.transition_counts))
            )



import pandas as pd
from tfgen.methods.classic import Classic
from threading import Thread
import tfgen.const as const
import queue

METHOD_CLASSIC = const.METHOD_CLASSIC


class TFGen:
    def __init__(self, observable_event_classes: list, window_size=500, method=METHOD_CLASSIC):
        """
        :param observable_event_classes: list of observable event classes. Use observe_event_classes to get the list if
        it's not available.
        :param window_size: The sliding window size.
        :param method:
        """

        self.ec_lookup = {ec: i for i, ec in enumerate(observable_event_classes)}

        # add predefined event classes
        self.ec_lookup[const.TOKEN_START_OF_TRACE] = len(self.ec_lookup)
        self.ec_lookup[const.TOKEN_END_OF_TRACE] = len(self.ec_lookup)
        self.ec_lookup[const.TOKEN_DEFAULT] = len(self.ec_lookup)

        self.window_size = window_size
        self.input_stream = queue.Queue(maxsize=10)
        self.output_stream = queue.Queue(maxsize=10)

        self._method_thread = Thread(target=self._select_method(method).start_processing)
        self._method_thread.start()

        self.input_method = 0

    def _select_method(self, method):
        if method == METHOD_CLASSIC:
            return Classic(self.ec_lookup, self.window_size, self.input_stream, self.output_stream)
        else:
            raise Exception("Method not supported")

    def _load_from_dataframe_thread(self, event_log, case_id_col, attributes_cols):
        case_ids = event_log[case_id_col].values
        attributes = event_log[attributes_cols].values
        """convert attributes to string"""

        samples = zip(case_ids, attributes)

        # _ = [self._input_stream.put(sample) for sample in samples]
        # _ = list(map(self._input_stream.put, samples))
        for sample in samples:
            self.input_stream.put(sample)

        # make finished signal
        self.input_stream.put((const.TOKEN_END_OF_TRACE, None))

    def load_from_dataframe(self, event_log: pd.DataFrame, case_id_col, attributes_cols):
        """
        Loads data from pandas dataframe. An offline process
        :param event_log: pandas dataframe
        :param case_id_col: the column name
        :param attributes_cols: array of column names
        """
        self.input_method = 0
        Thread(target=self._load_from_dataframe_thread, args=(event_log, case_id_col, attributes_cols)).start()

    def _load_from_generator_thread(self, generator):
        # put all data from the generator to the queue
        for sample in generator():
            self.input_stream.put(sample)

    def load_from_generator(self, generator):
        """
        Loads data from generator. Each yield is a tuple of (case_id, attributes), where attributes value is an iterable
        of strings (attribute_1, attribute_2, ...)
        """
        self.input_method = 1
        Thread(target=self._load_from_generator_thread, args=(generator,)).start()

    def load_next(self, case_id, attributes):
        """
        Load stream data one by one.
        :param case_id: str or int
        :param attributes: str. Iterable of attributes (attribute_1, attribute_2, ...).
        """
        self.input_method = 2
        self.input_stream.put((case_id, attributes))

    def get_output_generator(self):
        if self.input_method == 2:
            raise Exception("Cannot get generator with load_next() input."
                            "Use load_from_generator() or load_from_dataframe()")

        while True:
            data = self.output_stream.get()
            if data[0] == const.TOKEN_END_OF_TRACE:
                return

            yield data

    def quit(self):
        self.input_stream.put((const.TOKEN_END_OF_TRACE, None))

    def get_output_next(self):
        return self.output_stream.get(block=False)

    def get_output_list(self):
        """
        Returns a list of output samples.
        :param timeout: if the queue is empty for this amount of time, the process is
        assumed to be finished. Default is 1 second. Increase this value if your processing
        speed is less than 1 event/second.
        :return:
        """

        if self.input_method != 0:
            raise Exception("Cannot get offline output in online mode. Use get_output_generator() or "
                            "get_output_next() instead.")

        return list(self.get_output_generator())




from collections import deque

import pandas as pd
import numpy as np
from tfgen import const
from tfgen.models.tet import TET


class BaseMethod:
    def __init__(self, ec_lookup_table, window_size, input_stream, output_stream):
        self.ec_lookup_table = ec_lookup_table
        self.window_size = window_size
        self._input_stream = input_stream
        self._output_stream = output_stream

        self.tet = TET()
        self.sliding_window_buffer = deque(maxlen=self.window_size)

        self.event_log = None
        self.case_id_col = None
        self.attributes_cols = None

        self.current_case = None

        self.finished = False

        _len_lookup = len(ec_lookup_table)
        self.transition_counts = np.zeros((_len_lookup, _len_lookup), dtype=np.float)

    def index_lookup(self, ec):
        try:
            return self.ec_lookup_table[ec]
        except KeyError:
            return self.ec_lookup_table[const.TOKEN_DEFAULT]

    def process_a_single_event(self, data_tuple):
        raise NotImplementedError

    def start_processing(self):
        """
        Must return if self.finished is True.
        """
        raise NotImplementedError

    def get_next_data(self):
        data = self._input_stream.get()

        # end of file is reached
        if data[0] == const.TOKEN_END_OF_TRACE:
            self.put_data(data)
            self.finished = True
            return None
        else:
            return data

    def put_data(self, data):
        self._output_stream.put(data)



from setuptools import setup, find_packages
from os import path
from io import open

VERSION = '0.3.0'
DESCRIPTION = 'Online Transition-Based Feature Generation for Anomaly Detection in Concurrent Data Streams.'

current_directory = path.abspath(path.dirname(__file__))

with open(path.join(current_directory, 'README.rst'), encoding='utf-8') as f:
    readme = f.read()

with open(path.join(current_directory, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

setup(
    name="tfgen",
    version=VERSION,
    author="Yinzheng Zhong",
    author_email="y.zhong10@liverpool.ac.uk",
    url='https://github.com/yinzheng-zhong/TFGen',
    description=DESCRIPTION,
    long_description=readme,
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    install_requires=requirements,
    keywords=['anomaly detection', 'machine learning', 'intrusion detection', 'process mining'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: Education',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
    ]
)


import tfgen.const as const


class TET:
    def __init__(self):
        self.table = {}

    def update_event(self, case_id, event_class):
        if case_id not in self.table.keys():
            self.table[case_id] = event_class
            return const.TOKEN_START_OF_TRACE
        elif event_class == const.TOKEN_END_OF_TRACE:
            last_event = self.table[case_id]
            del self.table[case_id]
            return last_event
        elif case_id in self.table.keys():
            last_event = self.table[case_id]
            self.table[case_id] = event_class
            return last_event

